// C2 interface to the Cygnal (now Silicon Labs) processor.  This is used for
// flashing & debugging the Cygnal processor (C8051F3xx).
//
// Copyright 2005 Matt Otto <matt@klox.net>
// This file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.  It is distributed without any
// warranty, even the implied warranty of merchantability or fitness for a
// particular purpose.  See http://www.gnu.org/licenses/gpl.txt for the
// complete license.
//
//****************
// INCLUDES
//****************
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/types.h>
#include <time.h>
#include <linux/ioctl.h>
#include <assert.h>
#include "c2_ioctls.h"

//****************
// CONSTANTS
//****************

//#define UNUSED_C2_FUNCTIONS

// Mostly undocumented registers useful for flashing & debugging
// To access 0x00-0x02, I've only used C2_ReadReg & C2_WriteReg.
// To access 0x20-0x33, you must used C2_ReadRegBlock & C2_WriteRegBlock.
// To access 0x80-0xFF, you can use either method.
#define DEVICEID 0x00   // fixed device ID value (should match DevIDList)
#define REVID    0x01   // fixed revision ID (I've only seen 0x00 on an F331)
#define FPCTL    0x02   // Flash Programming Control (also for debugging)
#define PCL_COPY 0x20   // copy of Program Counter, low byte, while debugging
#define PCH_COPY 0x21   // copy of Program Counter, high byte, while debugging
#define PSW_COPY 0x23   // copy of PSW while debugging
#define R0_COPY  0x24   // copy of R0 (bank 0) while debugging
#define R1_COPY  0x25   // copy of R1 (bank 0) while debugging
#define R2_COPY  0x26   // copy of R2 (bank 0) while debugging
#define XRAMD    0x84   // XRAM Data (address auto-increments)
#define BRKP0L   0x85   // Breakpoint 0, low byte (write only)
#define BRKP0H   0x86   // Breakpoint 0, high byte (write only)
#define BRKP1L   0xAB   // Breakpoint 1, low byte (write only)
#define BRKP1H   0xAC   // Breakpoint 1, high byte (write only)
#define XRAMAL   0xAD   // XRAM Address, low byte (address auto-increments)
#define FPDAT    0xB4   // Flash Programming Data (also for debugging)
#define XRAMAH   0xC7   // XRAM Address, high byte (address auto-increments)
#define BRKP2L   0xCE   // Breakpoint 2, low byte (write only)
#define BRKP2H   0xCF   // Breakpoint 2, high byte (write only)
#define PSW      0xD0   // Program Status Word (this is a documented register)
#define BRKP3L   0xD2   // Breakpoint 3, low byte (write only)
#define BRKP3H   0xD3   // Breakpoint 3, high byte (write only)

// FPCTL codes
#define INIT_KEY1   0x02        // first key
#define INIT_KEY2   0x01        // second key
#define RESUME_EXEC 0x08        // resume code execution

// FPDAT commands
#define UNKNOWN1     0x01
#define UNKNOWN2     0x02
#define DEVICE_ERASE 0x03
#define FLASH_READ   0x06
#define FLASH_WRITE  0x07
#define PAGE_ERASE   0x08
#define REG_READ     0x09
#define REG_WRITE    0x0A
#define RAM_READ     0x0B
#define RAM_WRITE    0x0C

// FPDAT status codes
#define INVALID_COMMAND 0x00
#define COMMAND_FAILED  0x02
#define COMMAND_OK      0x0D

// FPI Command Status (C2_ReadAddr) bit masks
#define FLASHING       0x80   // this isn't on continuously during Device Erase
#define BREAKPOINT_HIT 0x04   // after stopping due to a breakpoint, this stays on until the next RESUME_EXEC
#define INBUSY         0x02
#define OUTREADY       0x01

// The worst-case wait should be during a Flash Device Erase, which can take
// up to 20ms per page.  For 16kB (which is the largest C2-based MCU as of
// this writing), that's about 640ms.  Since the WaitFor* routines get
// called often, I use the time() function because it's simple.  The shortest
// timeout using time() is 1-2 seconds, which is good enough.
#define POLL_TIMEOUT 2    // 2 = at most 2 seconds, and as little as 1 second

//****************
// MACROS
//****************

#define min(a,b)  (((a) < (b)) ? (a) : (b))
#define max(a,b)  (((a) > (b)) ? (a) : (b))

//****************
// VARIABLES
//****************

// A quick register address lookup for when setting breakpoints
__u8 BPTable[4][2] = {
  {BRKP0L, BRKP0H},
  {BRKP1L, BRKP1H},
  {BRKP2L, BRKP2H},
  {BRKP3L, BRKP3H}
};

// Expected value in DEVICEID, according to the datasheets
struct {
  __u8 ID;
  char NameString[20];
} DevIDList[] = {
  {0x04, "F30x"},
  {0x09, "F31x or F32x"},
  {0x0A, "F33x"},
  {0x0B, "F35x"}
};
#define DEVIDLIST_LENGTH 4

int driver_handle;  // handle to the cygnal_c2 driver

//*********************************************************************

//****************
// C2_ReadData
//****************
inline __u8 C2_ReadData(void)
{
  unsigned long tmp;
  ioctl(driver_handle, C2_IOCD1READ, &tmp);
  return (__u8)tmp;
}

//****************
// C2_WriteData
//****************
inline void C2_WriteData(__u8 dat)
{
  ioctl(driver_handle, C2_IOCD1WRITE, &dat);
}

//****************
// C2_ReadAddr
//****************
inline __u8 C2_ReadAddr(void)
{
  unsigned long tmp;
  ioctl(driver_handle, C2_IOCAREAD, &tmp);
  return (__u8)tmp;
}

//****************
// C2_WriteAddr
//****************
inline void C2_WriteAddr(__u8 addr)
{
  ioctl(driver_handle, C2_IOCAWRITE, &addr);
}

//****************
// C2_ReadReg
//****************
__u8 C2_ReadReg(__u8 reg)
{
  C2_WriteAddr(reg);
  return C2_ReadData();
}

//****************
// C2_WriteReg
//****************
void C2_WriteReg(__u8 reg, __u8 val)
{
  C2_WriteAddr(reg);
  C2_WriteData(val);
}

//****************
// WaitForOutReady
//****************
inline void WaitForOutReady(void)
{
  time_t limit = time(NULL) + POLL_TIMEOUT;

  while (!(C2_ReadAddr() & OUTREADY)) {
    assert(time(NULL) < limit);
  }
}
//****************
// WaitForInReady
//****************
inline void WaitForInReady(void)
{
  // "InReady" means the InBusy bit is NOT set

  time_t limit = time(NULL) + POLL_TIMEOUT;

  while (C2_ReadAddr() & INBUSY) {
    assert(time(NULL) < limit);
  }
}

//*********************************************************************

//****************
// C2_StopMCU
//****************
inline void C2_StopMCU(void)
{
  ioctl(driver_handle, C2_IOCQRESET, NULL);    // stop execution on the MCU
}

//****************
// C2_ResumeMCU
//****************
inline void C2_ResumeMCU(void)
{
  C2_WriteReg(FPCTL,RESUME_EXEC);            // resume execution on the MCU
}

//****************
// C2_SetBreakpoint
//****************
void C2_SetBreakpoint(__u8 bp_num, __u16 Addr)
{
  C2_WriteReg(BPTable[bp_num][0], (__u8)Addr);
  C2_WriteReg(BPTable[bp_num][1], (__u8)(Addr >> 8) | 0x80);
}

//****************
// C2_ClearBreakpoint
//****************
void C2_ClearBreakpoint(__u8 bp_num)
{
  C2_WriteReg(BPTable[bp_num][1], 0x00);
}

//****************
// C2_CheckForBreakpoint
//****************
inline int C2_CheckForBreakpoint(void)
{
  return ioctl(driver_handle, C2_IOCIRQCHK, NULL);
}

//****************
// C2_WaitForBreakpoint
//****************
inline void C2_WaitForBreakpoint(void)
{
  // Keep checking for an interrupt from the MCU,
  // which indicates that the breakpoint was hit.
  while (C2_CheckForBreakpoint() == 0);
}

#ifdef UNUSED_C2_FUNCTIONS
//****************
// C2_UnknownCmd1
//****************
__u8 C2_UnknownCmd1(void)
{
  // I don't know what ths FPDAT command does, but it reads 1 byte
  // Returns the byte read, or 0xFF on error

  C2_WriteReg(FPDAT, UNKNOWN1);  // load the command into FPDAT
  WaitForInReady();              // Wait for input acknowledge

  WaitForOutReady();             // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 0xFF;                 // error
  }

  WaitForOutReady();             // Wait for the unknown byte
  return C2_ReadData();          // return the unknown byte
}

//****************
// C2_UnknownCmd2
//****************
__u8 C2_UnknownCmd2(void)
{
  // I don't know what ths FPDAT command does, but it reads 1 byte
  // Returns the byte read, or 0xFF on error

  C2_WriteReg(FPDAT, UNKNOWN2);  // load the command into FPDAT
  WaitForInReady();              // Wait for input acknowledge

  WaitForOutReady();             // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 0xFF;                 // error
  }

  WaitForOutReady();             // Wait for the unknown byte
  return C2_ReadData();          // return the unknown byte
}
#endif  // UNUSED_C2_FUNCTIONS

//****************
// C2_DeviceErase
//****************
__u8 C2_DeviceErase(void)
{
  // Erase the entire Flash memory space.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(FPDAT, DEVICE_ERASE); // load the DEVICE_ERASE command into FPDAT
  WaitForInReady();                 // Wait for input acknowledge

  // Check status before starting Flash access sequence
  WaitForOutReady();                // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                       // error
  }

  // Send a three-byte arming sequence to enable the device erase. If the
  // sequence is not received correctly, the command will be ignored.
  // Sequence: 0xDE, 0xAD, 0xA5.
  C2_WriteData(0xDE);           // Arming sequence command 1
  WaitForInReady();             // Wait for input acknowledge

  C2_WriteData(0xAD);           // Arming sequence command 2
  WaitForInReady();             // Wait for input acknowledge

  C2_WriteData(0xA5);           // Arming sequence command 3
  WaitForInReady();             // Wait for input acknowledge

  WaitForOutReady();            // Wait for erase operation to complete
  return 0;                     // success
}

//****************
// C2_ReadFlashBlock
//****************
__u8 C2_ReadFlashBlock(__u16 BlockStart, __u8 BlockSize, __u8 *BufPtr)
{
  // Read a block of Flash starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(FPDAT, FLASH_READ);    // load the FLASH_READ command into FPDAT
  WaitForInReady();                  // Wait for input acknowledge

  // Check status before starting Flash access sequence
  WaitForOutReady();                 // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                        // error
  }

  C2_WriteData(BlockStart >> 8);     // Send address high byte to FPDAT
  WaitForInReady();                  // Wait for input acknowledge
  C2_WriteData(BlockStart & 0x00FF); // Send address low byte to FPDAT
  WaitForInReady();                  // Wait for input acknowledge
  C2_WriteData(BlockSize);           // Send block size
  WaitForInReady();                  // Wait for input acknowledge

  // Check status before reading Flash block
  WaitForOutReady();                 // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                        // error
  }

  // Read Flash block (BlkSize of 0x00 = 256)
  do
  {
    WaitForOutReady();            // Wait for data ready indicator
    *BufPtr++ = C2_ReadData();    // Read data from the FPDAT register
  } while (--BlockSize);

  return 0;                       // success
}

//****************
// C2_WriteFlashBlock
//****************
__u8 C2_WriteFlashBlock(__u16 BlockStart, __u8 BlockSize, __u8 *BufPtr)
{
  // Write a block of Flash starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(FPDAT, FLASH_WRITE);  // load the FLASH_WRITE command into FPDAT
  WaitForInReady();                 // Wait for input acknowledge

  // Check status before starting Flash access sequence
  WaitForOutReady();                  // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                         // error
  }

  C2_WriteData(BlockStart >> 8);      // Send address high byte to FPDAT
  WaitForInReady();                   // Wait for input acknowledge
  C2_WriteData(BlockStart & 0x00FF);  // Send address low byte to FPDAT
  WaitForInReady();                   // Wait for input acknowledge
  C2_WriteData(BlockSize);            // Send block size
  WaitForInReady();                   // Wait for input acknolwedge

  // Check status before writing Flash block
  WaitForOutReady();                  // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                         // error
  }

  // Write Flash block (BlkSize of 0x00 = 256)
  do
  {
    C2_WriteData(*BufPtr++);          // Write data to the FPDAT register
    WaitForInReady();                 // Wait for input acknowledge
  } while (--BlockSize);

  WaitForOutReady();                  // Wait for last Flash write to complete
  return 0;                           // success
}

//****************
// C2_FlashPageErase
//****************
__u8 C2_FlashPageErase(__u16 BlockStart)
{
  // Erase a 512-byte Flash page starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  __u8 page;                         // Target Flash page

  C2_WriteReg(FPDAT, PAGE_ERASE);  // load the PAGE_ERASE command into FPDAT
  WaitForInReady();                // Wait for input acknowledge

  // Check status before starting Flash access sequence
  WaitForOutReady();               // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                      // error
  }

  // page is the 512-byte sector containing the target BlockStart.
  page = (__u8)(BlockStart >> 9);

  C2_WriteData(page);              // Send Flash page number
  WaitForInReady();                // Wait for input acknowledge

  WaitForOutReady();               // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                      // error
  }

  C2_WriteData(0x00);              // Dummy write to initiate erase
  WaitForInReady();                // Wait for input acknowledge

  WaitForOutReady();               // Wait for erase operation to complete
  return 0;                        // success
}

//****************
// C2_ReadRegBlock
//****************
__u8 C2_ReadRegBlock(__u8 BlockStart, __u8 BlockSize, __u8 *BufPtr)
{
  // Reads a block of registers starting at BlockStart.
  // Returns 0 if successful, 1 otherwise
 
  C2_WriteReg(FPDAT, REG_READ);    // load the REG_READ command into FPDAT
  WaitForInReady();                // Wait for input acknowledge

  // Check status before starting register access sequence
  WaitForOutReady();               // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                      // error
  }

  C2_WriteData(BlockStart);        // Send address byte to FPDAT
  WaitForInReady();                // Wait for input acknowledge
  C2_WriteData(BlockSize);         // Send block size
  WaitForInReady();                // Wait for input acknowledge

  do
  {
    WaitForOutReady();             // Wait for data ready indicator
    *BufPtr++ = C2_ReadData();     // Read data from the FPDAT register
  } while (--BlockSize);

  return 0;                        // success
}

//****************
// C2_WriteRegBlock
//****************
__u8 C2_WriteRegBlock(__u8 BlockStart, __u8 BlockSize, __u8 *BufPtr)
{
  // Writes a block of registers starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(FPDAT, REG_WRITE);   // load the REG_WRITE command into FPDAT
  WaitForInReady();                // Wait for input acknowledge

  // Check status before starting register access sequence
  WaitForOutReady();               // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                      // error
  }

  C2_WriteData(BlockStart);        // Send address byte to FPDAT
  WaitForInReady();                // Wait for input acknowledge
  C2_WriteData(BlockSize);         // Send block size
  WaitForInReady();                // Wait for input acknowledge

  do
  {
    C2_WriteData(*BufPtr++);       // Write data to the FPDAT register
    WaitForInReady();              // Wait for input acknowledge
  } while (--BlockSize);

  return 0;                        // success
}

//****************
// C2_ReadRAMBlock
//****************
__u8 C2_ReadRAMBlock(__u8 BlockStart, __u8 BlockSize, __u8 *BufPtr)
{
  // Reads a block of internal RAM starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(FPDAT, RAM_READ);    // load the RAM_READ command into FPDAT
  WaitForInReady();                // Wait for input acknowledge

  // Check status before starting RAM access sequence
  WaitForOutReady();               // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                      // error
  }

  C2_WriteData(BlockStart);        // Send address byte to FPDAT
  WaitForInReady();                // Wait for input acknowledge
  C2_WriteData(BlockSize);         // Send block size
  WaitForInReady();                // Wait for input acknowledge

  do
  {
    WaitForOutReady();             // Wait for data ready indicator
    *BufPtr++ = C2_ReadData();     // Read data from the FPDAT register
  } while (--BlockSize);

  return 0;                        // success
}

//****************
// C2_WriteRAMBlock
//****************
__u8 C2_WriteRAMBlock(__u8 BlockStart, __u8 BlockSize, __u8 *BufPtr)
{
  // Writes a block of internal RAM starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(FPDAT, RAM_WRITE);   // load the RAM_WRITE command into FPDAT
  WaitForInReady();                // Wait for input acknowledge

  // Check status before starting RAM access sequence
  WaitForOutReady();               // Wait for status
  if (C2_ReadData() != COMMAND_OK)
  {
    return 1;                      // error
  }

  C2_WriteData(BlockStart);        // Send address byte to FPDAT
  WaitForInReady();                // Wait for input acknowledge
  C2_WriteData(BlockSize);         // Send block size
  WaitForInReady();                // Wait for input acknowledge

  do
  {
    C2_WriteData(*BufPtr++);       // Write data to the FPDAT register
    WaitForInReady();              // Wait for input acknowledge
  } while (--BlockSize);

  return 0;                        // success
}

//****************
// C2_ReadXRAMBlock
//****************
__u8 C2_ReadXRAMBlock(__u16 BlockStart, __u16 BlockSize, __u8 *BufPtr)
{
  // Reads a block of XRAM starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(XRAMAL, (__u8)BlockStart);          // load the address
  C2_WriteReg(XRAMAH, (__u8)(BlockStart >> 8));

  // read repeatedly from XRAMD, the address in XRAMA will auto-increment
  C2_WriteAddr(XRAMD);
  do
  {
    *BufPtr++ = C2_ReadData();     // read data
  } while (--BlockSize);

  return 0;                        // success
}

//****************
// C2_WriteXRAMBlock
//****************
__u8 C2_WriteXRAMBlock(__u16 BlockStart, __u16 BlockSize, __u8 *BufPtr)
{
  // Writes a block of XRAM starting at BlockStart.
  // Returns 0 if successful, 1 otherwise

  C2_WriteReg(XRAMAL, (__u8)BlockStart);          // load the address
  C2_WriteReg(XRAMAH, (__u8)(BlockStart >> 8));

  // write repeatedly from XRAMD, the address in XRAMA will auto-increment
  C2_WriteAddr(XRAMD);
  do
  {
    C2_WriteData(*BufPtr++);       // write data
  } while (--BlockSize);

  return 0;                        // success
}

//*********************************************************************

//****************
// C2_InitializeInterface
//****************
__u8 C2_InitializeInterface(void)
{
  struct timespec tspec;

  ioctl(driver_handle,C2_IOCRESET,NULL);   // reset target device

  tspec.tv_sec = 0;
  tspec.tv_nsec = 2000L;
  nanosleep(&tspec, NULL);   // Delay for at least 2us

  // Initialize the C2 Interface for flashing & debugging

  C2_WriteAddr(FPCTL);       // Target the C2 Flash Programming Control
                             // register (FPCTL) for C2 Data register accesses

  C2_WriteData(INIT_KEY1);   // Write the first key code to enable C2 control
  C2_WriteData(INIT_KEY2);   // Write the second key code to enable C2 control

  tspec.tv_sec = 0;
  tspec.tv_nsec = 20000000L; // Delay for at least 20ms to ensure the
  nanosleep(&tspec, NULL);

  // target is ready for Debugging / Flash programming

  //C2_WriteReg(FPCTL,RESUME_EXEC);   // resume execution on the MCU

  return 0;                         // success
}

//****************
// C2_FlashCygnal
//****************
__u8 C2_FlashCygnal(__u8 *FlashBuffer, __u16 BufferSize)
{
  // Returns 0 if successful, 1 otherwise

  __u16 index, BlockSize;
  __u8 *BufPtr;

  // Erase entire Flash code space
  if (C2_DeviceErase())
  {
    return 1;
  }

  for (index=0, BufPtr=FlashBuffer; index<BufferSize; index+=BlockSize, BufPtr+=BlockSize)
  {
    // read 256-byte blocks except for the last one
    BlockSize = min(BufferSize-index, 256);

    // by casting BlockSize to a __u8, 256 becomes 0, which is intentional
    if (C2_WriteFlashBlock(index, (__u8)BlockSize, BufPtr))
    {
      return 1;
    }
  }

  return 0;
}

//*********************************************************************

// 0x20 is arbitrary: it could be as big as 0x100, but if you change it,
// fix up the code to read a fractional block as the last interation
// through each loop.
#define BUF_SZ 0x20

//****************
// C2_CoreDump
//****************
__u8 C2_CoreDump(void)
{
  // Returns 0 if successful, 1 otherwise

  __u8 buf[BUF_SZ], debug_buf[0x14];
  __u16 byte_cnt, blk_cnt;

  printf("DEVICEID = 0x%02X\n", C2_ReadReg(DEVICEID));
  printf("REVID = 0x%02X\n", C2_ReadReg(REVID));
  // There's no reason to dump FPCTL.

  // Read 0x20-0x33 so we have all the register copies,
  // plus whatever else is in there.
  if (C2_ReadRegBlock(0x20, 0x14, debug_buf))
  {
    printf("Reg read at 0x20 failed\n");
    return 1;
  }
  printf("PC = 0x%02X%02X\n", debug_buf[1], debug_buf[0]);

#if 0  // Is there a reason to display the rest?
  printf("Debug = \n");
  for (byte_cnt = 0; byte_cnt < 0x14;)
  {
    // start the line with the address
    if ((byte_cnt & 0x0F) == 0x00)
    {
      printf("0x%02X:", 0x20 + byte_cnt);
    }
    printf(" %02X", debug_buf[byte_cnt]);
    // Every 8 bytes, add a space.  Every 16 bytes, start a new line.
    switch (++byte_cnt & 0x0F)
    {
    case 0x08:
      printf(" ");
      break;
    case 0x00:
      printf("\n");
      break;
    }
  }
  printf("\n");
#endif

  printf("SFRs = \n");
  for (blk_cnt = 0x80; blk_cnt < 0x100; blk_cnt += BUF_SZ)
  {
    if (C2_ReadRegBlock(blk_cnt, BUF_SZ, buf))
    {
      printf("Reg read at 0x%02X failed\n", blk_cnt);
      return 1;
    }
    for (byte_cnt = 0; byte_cnt < BUF_SZ;)
    {
      // start the line with the address
      if (((byte_cnt + blk_cnt) & 0x0F) == 0x00)
      {
        printf("0x%02X:", byte_cnt + blk_cnt);
      }
      // we want our copy of PSW, not the C2 ROM Monitor's copy
      if ((blk_cnt + byte_cnt) == PSW)
      {
        printf(" %02X", debug_buf[0x03]);   // PSW_COPY is at 0x23
      }
      else
      {
        printf(" %02X", buf[byte_cnt]);
      }
      // Every 8 bytes, add a space.  Every 16 bytes, start a new line.
      switch ((++byte_cnt + blk_cnt) & 0x0F)
      {
      case 0x08:
        printf(" ");
        break;
      case 0x00:
        printf("\n");
        break;
      }
    }
  }

  printf("RAM = \n");
  for (blk_cnt = 0; blk_cnt < 0x100; blk_cnt += BUF_SZ)
  {
    if (C2_ReadRAMBlock(blk_cnt, BUF_SZ, buf))
    {
      printf("RAM read at 0x%02X failed\n", blk_cnt);
      return 1;            // error
    }
    for (byte_cnt = 0; byte_cnt < BUF_SZ;)
    {
      // start the line with the address
      if (((byte_cnt + blk_cnt) & 0x0F) == 0x00)
      {
        printf("0x%02X:", byte_cnt + blk_cnt);
      }
      // we want our copy of R0-R2, not the C2 ROM Monitor's copy
      switch (blk_cnt + byte_cnt)
      {
      case 0x00:
        printf(" %02X", debug_buf[0x04]);   // R0_COPY is at 0x24
        break;
      case 0x01:
        printf(" %02X", debug_buf[0x05]);   // R1_COPY is at 0x25
        break;
      case 0x02:
        printf(" %02X", debug_buf[0x06]);   // R2_COPY is at 0x26
        break;
      default:
        printf(" %02X", buf[byte_cnt]);
        break;
      }
      // Every 8 bytes, add a space.  Every 16 bytes, start a new line.
      switch ((++byte_cnt + blk_cnt) & 0x0F)
      {
      case 0x08:
        printf(" ");
        break;
      case 0x00:
        printf("\n");
        break;
      }
    }
  }

#if 0  // I don't care about XRAM and flash most of the time (too much data!)
  printf("XRAM = \n");
  for (blk_cnt = 0; blk_cnt < 0x200; blk_cnt += BUF_SZ)
  {
    if (C2_ReadXRAMBlock(blk_cnt, BUF_SZ, buf))
    {
      printf("XRAM read at 0x%02X failed\n", blk_cnt);
      return 1;            // error
    }
    for (byte_cnt = 0; byte_cnt < BUF_SZ;)
    {
      // start the line with the address
      if (((byte_cnt + blk_cnt) & 0x0F) == 0x00)
      {
        printf("0x%03X:", byte_cnt + blk_cnt);
      }
      printf(" %02X", buf[byte_cnt]);
      // Every 8 bytes, add a space.  Every 16 bytes, start a new line.
      switch ((++byte_cnt + blk_cnt) & 0x0F)
      {
      case 0x08:
        printf(" ");
        break;
      case 0x00:
        printf("\n");
        break;
      }
    }
  }

  //FIXME: This assumes that flash is 8192-512 bytes.  Not all MCUs are 8kB.
  printf("Flash = \n");
  for (blk_cnt = 0; blk_cnt < (0x2000-0x200); blk_cnt += BUF_SZ)
  {
    if (C2_ReadFlashBlock(blk_cnt, BUF_SZ, buf))
    {
      if ((blk_cnt & 0x01FF) == 0x0000)   // Is this the beginning of a page?
      {
        // Probably a locked page, skip this page and try again.
        printf("0x%04X: Page locked", blk_cnt);
        blk_cnt += 512-BUF_SZ;
        continue;
      }
      printf("flash read at 0x%02X failed\n", blk_cnt);
      return 1;               // Not a locked page.  Unexpected error.
    }
    for (byte_cnt = 0; byte_cnt < BUF_SZ;)
    {
      // start the line with the address
      if (((byte_cnt + blk_cnt) & 0x0F) == 0x00)
      {
        printf("0x%04X:", byte_cnt + blk_cnt);
      }
      printf(" %02X", buf[byte_cnt]);
      // Every 8 bytes, add a space.  Every 16 bytes, start a new line.
      switch ((++byte_cnt + blk_cnt) & 0x0F)
      {
      case 0x08:
        printf(" ");
        break;
      case 0x00:
        printf("\n");
        break;
      }
    }
  }
#endif

  return 0;              // success
}

int main(int argc, char * argv[])
{
  __u8 ret_val, index;

  driver_handle = open("cygnal", O_RDWR);
  if (driver_handle < 0) {
    printf("Can't open the \"cygnal\" special file\n");
    return 1;
  }

  C2_InitializeInterface();

  ret_val = C2_ReadReg(DEVICEID);
  for (index=0; index < DEVIDLIST_LENGTH; index++) {
    if (DevIDList[index].ID == ret_val) {
      printf("%s detected.\n", DevIDList[index].NameString);
      break;
    }
  }
  if (index == DEVIDLIST_LENGTH) {
    printf("C2 initialization failed. Unknown DevID 0x%02X.\n", (int)ret_val);
    return 1;
  }

  //
  // The rest of this function is testing code
  //

#if 1
  // Looking at the *.rst output from sdcc, I know that my MCU code will
  // hit address 0x008E.  Let's set a breakpoint and wait for it to hit it.
  C2_SetBreakpoint(0, 0x008E);
  C2_CheckForBreakpoint();   // clear out any old interrupts
  C2_ResumeMCU();
  C2_WaitForBreakpoint();
  C2_ClearBreakpoint(0);
#endif

#if 1
  // Instead of using a breakpoint, let's just run the MCU for
  // 2 seconds and then see where it's at.
  C2_ResumeMCU();
  sleep(2);
  C2_StopMCU();
#endif
  ret_val = C2_CoreDump();

  close(driver_handle);
  return (int)ret_val;
}

